#
#

package Walker;

require Retriever;
require HTTPRetriever;
require FTPRetriever;
require GopherRetriever;

require CHandler;
require CompressHandler;
require DirectoryHandler;
require DocumentHandler;
require FortranHandler;
require HTMLHandler;
require ImageHandler;
require LISPHandler;
require MiscHandler;
require PerlHandler;
require SoundHandler;
require TarHandler;
require TkHandler;
require IndexingHandler;

require URI::URL;

#
# /home/mmlab1/Software/CVSMaster/walker/Walker.pm,v 1.8 1996/01/18 17:48:59 olson Exp
#
# Walker.pm,v
# Revision 1.8  1996/01/18 17:48:59  olson
# added rcs headers
#
#
#
# Object: Walker
#
#
# Instance variables:

#	OBJ_config_file_entries
#
#	retriever_table
#
#		Hash mapping from protocol name to retrieval object
#
#	local_queue
#
#		Local list of URLs to be searched.
#
# Methods
#
#	enqueue_url
#
#		Add a url to the local queue.
#
#	

@ISA = qw(Object);

#
# Register subroutines with the walker process.
#
sub boot
{
    Nexus::register_handler("walk_url", "walk_url",
			    method_call => 1, unpack_buffer => 1);
    Nexus::register_handler("start", "start",
			    method_call => 1, unpack_buffer => 1);
    Nexus::register_handler("get_stats", "get_stats",
			    method_call => 1, unpack_buffer => 1);
}

#
# Create a new walker process.
#
sub new
{
    my($class) = @_;
    my $self = {};
    bless $self, $class;
    $$self{'OBJ_config_file_entries'} = {};

#    print "Walker new\n";

    $self->generate_accessors('hdb',
			      'mdb',
			      'id',
			      'pid',
			      'local_queue',
			      'handlers',
			      'walker',
			      'num_pages',
			      'tot_size',
			      'default_handler',
			      'handler_cache',
			      'handler_suffix_cache',
			      'retriever_table');

    $self->set_retriever_table({'http'    => new HTTPRetriever,
				'ftp'     => new FTPRetriever,
				'gopher'  => new GopherRetriever,
				});

    $self->set_handler_cache({});
    $self->set_handler_suffix_cache({});

    $self->set_local_queue([]);

    return $self;
}

#
# Start up the object.
#
sub start
{
    my($self) = @_;
    my($hdb);
    my($url);
    my($queue);
    my($handlers);

    $| = 1;

    print "WALKER ", $self->id(), ": starting\n";

#    open(STDOUT, ">walker.out." . $self->id);
#    open(STDERR, ">&STDOUT");
#    select(STDERR);
#    $| = 1;

#    select(STDOUT);
#    $| = 1;

    #
    # Create the handlers here as we know here that the mdb has been set.
    #
    $handlers = [];
#    push(@$handlers, CHandler->new($self->mdb));
#    push(@$handlers, CompressHandler->new($self->mdb));
#    push(@$handlers, DirectoryHandler->new($self->mdb));
#    push(@$handlers, FortranHandler->new($self->mdb));
    push(@$handlers, HTMLHandler->new($self->mdb));
#    push(@$handlers, ImageHandler->new($self->mdb));
#    push(@$handlers, LISPHandler->new($self->mdb));
#    push(@$handlers, PerlHandler->new($self->mdb));
#    push(@$handlers, SoundHandler->new($self->mdb));
#    push(@$handlers, TarHandler->new($self->mdb, $self));
#    push(@$handlers, TkHandler->new($self->mdb));

    my $ih = IndexingHandler->new($self->mdb);
    if ($ih)
    {
	push(@$handlers, $ih);
    }
    $self->set_handlers($handlers);

    $self->set_default_handler(MiscHandler->new($self->mdb));

    $self->set_num_pages(0);
    $self->set_tot_size(0);

    $hdb = $self->hdb;

    $queue = $self->local_queue;

#    print "Invoke $hdb/get_url with $self\n";
    Nexus::rpc_async($hdb, "get_url", $self, "walk_url", $self->id());
}

#
# "Walk" the walker by retrieving a url.
#
sub walk_url
{
    my($self, $url_string) = @_;
    my($scheme,$host,$port,$path,$query,$frag);
    my($me);

    $url = new URI::URL $url_string;

    if ($url)
    {
	
	$scheme = $url->scheme;
	$host = $url->host;
	$port = $url->port;
	$path = $url->path;
	$query = $url->query;
	$frag = $url->frag;
	
	$retriever = $self->find_retriever($scheme);
	
	if ($retriever)
	{
	    $me = "Walker " . $self->id();
#	    print uc($me), ": retrieving $url with $retriever\n";
	    $0 = "Walker ($me $url)";
	    $self->retrieve_url($url, $retriever);
	    $0 = "Walker ($me)";
#	    print uc($me), ": retrieved $url with $retriever\n";
	}
    }
    my $hdb = $self->hdb;
    print "Invoke $hdb/get_url with $self ", $self->id(), "\n";
    Nexus::rpc_async($hdb, "get_url", $self, "walk_url", $self->id());
}

#
# Retrieve a page from the Web.
#
sub retrieve_url
{
    my($self, $url_string, $retriever) = @_;
    my($url_loc, $page_loc, $url, $p, @p);
    my(@h, $h, $host, $path);

    $url = new URI::URL $url_string;

    return unless defined($url);

    #
    # Ergh. I'd rather these stayed my(), but the libwwww
    # wants to get things passed as globs.
    #
    local(%headers, $content, $respcode, $headers);

    my $content_glob = *content;

    $respcode = $retriever->get($url->as_string, $content_glob, *headers);

    print "Retrieve returns $respcode\n" if $debug;

    $num_pages = $self->num_pages();
    $num_pages++;
    $self->set_num_pages($num_pages);
    $tot_size = $self->tot_size();
    $tot_size += length($content);
    $self->set_tot_size($tot_size);

    if ($respcode !~ /^2/)
    {
	print "Didn't get page $url\n" if $debug;
	return;
    }

    $host = $url->host;
    $path = $url->path;

    @h = reverse(split(/\./, $host));

    $page_loc = "$::page_dir";
     
    #
    # Save the Web page on local disk. We do this here, and
    # make the location available to the handlers if they
    # want the actual content.
    #
    for $h (@h)
    {
	$page_loc .= '/' . $h;
	if (! -d $page_loc)
	{
	    if (!mkdir($page_loc, 0775))
            {
	        warn "Cannot mkdir $page_loc: $!\n";
		return '/tmp';
	    }
	}
    }

    $page_loc .= "/";

    if ($path)
    {
	@p = split(/\//, $path);
	$file = pop(@p);
        for $p (@p)
	{
	    $page_loc .= "$p";
	    if ((-e $page_loc) && (! -d $page_loc))
	    {
#		print "fixing $page_loc\n";
		rename($page_loc, "$page_loc.index") ||
		warn "rename $page_loc $page_loc.index failed: $!\n";
		mkdir($page_loc, 0755) || warn "mkdir $page_loc failed: $!\n";
		rename("$page_loc.index", "$page_loc/index.html") ||
		warn "rename $page_loc.index $page_loc/index.html failed: $!\n";
            }
	    elsif (! -e $page_loc)
	    {
		mkdir($page_loc, 0755) || warn "mkdir $page_loc failed: $!\n";
            }
        }
    }

    if (! $file)
    {
	$file = "index.html";
    }
    
    $url_loc = "$page_loc/$file";
    if (open(OUT, ">$url_loc"))
    {
        print OUT $content;
        close(OUT);
    }
    else
    {
	warn "Cannot open $url_loc: $!\n";
    }

    $self->get_handler($url, $content_glob, $url_loc, %headers);

    #unlink($url_loc);
}

#
# Get a handler to deal with the Web page.
#
# Not really "get" a handler, rather handle the page.
#
sub get_handler
{
    my($self, $url, $content_glob, $url_loc, %headers) = @_;
    my($handler_listref, $handler);

    $handler_listref = $self->find_handlers($url, $headers{'content-type'},
					    $content_glob, $url_loc);

    for $handler (@$handler_listref)
    {
	print "Handling $url with $handler\n" if $debug;
	$handler->handle($url, $content_glob, *headers, $url_loc);
    }
}

#
# Find the best handler to deal with the Web page.  Index the page if it is a
# text page.
#
# Modified to return a list of handlers.  RDO 1/10/96
#
sub find_handlers
{
    my($self, $url, $content_type, $content, $url_loc) = @_;
    my($handler_listref, $suffix, $handler);

    #
    # Removed code here that used the got_image and got_sound
    # handler methods. The indexing handler will know whether
    # it wants the type or not.
    #

    ($suffix) = ($url =~ /\.([^.]+)$/);

    #
    # Find a handler for this document
    #
    # If we don't have a document type, or if it is text/plain, take
    # a look at its suffix. Pick a handler based on the suffix.
    #
    # Arg, this is too complex.
    #
    # If we don't have a document type, and we do have a suffix,
    # fake the document type to be text/suffix-<suffix>
    #
    # If it doesn't have a suffix, bail & call it text/plain.
    #
    #
    if ($content_type eq '')
    {
	if ($suffix eq '')
	{
	    $content_type = "text/plain";
	}
	else
	{
	    $content_type = "text/suffix-$suffix";
	}
    }
    
    #
    # Make the handler cache contain a ref to a list
    # of handlers. This way multiple handlers can handle each
    # document type.
    #
    if (defined($handler_listref = $self->handler_cache->{$content_type}))
    {
	return $handler_listref;
    }
    
    $handler_listref = [];
    
    for $handler (@{$self->handlers})
    {
	if ($handler->accepts($content_type))
	{
	    push(@$handler_listref, $handler);
	    #$self->handler_cache->{$content_type} = $handler;
	    #return $handler;
	}
    }
    if (@$handler_listref == 0)
    {
	$handler = $self->default_handler;
	push(@$handler_listref, $handler);
	print "Defaulting to MiscHandler for content $content_type for $url\n" if $debug;
    }

    $self->handler_cache->{$content_type} = $handler_listref;
    return $handler_listref;
}

#
# Find the correct retriever for the current url.
#
sub find_retriever
{
    my($self, $scheme) = @_;
    my($obj);

    if (!defined($obj = $self->retriever_table->{$scheme}))
    {
	warn "No retriever defined for $scheme\n";
	return undef;
    }
    else
    {
	return $obj;
    }
}
  
#
# Get the statistics for the current walker process.
#
sub get_stats
{
    my($self, $return_gp) = @_;
    my($a);
    $a = [$self->num_pages(), $self->tot_size()];
    return $a;
}

sub DESTROY
{
    print "HEY SOMEONE'S DESTROYING @_\n";
}

1;
