#
#

package HDB;
use Fcntl;

$idx = 34;

#
# /home/mmlab1/Software/CVSMaster/walker/hdb/HDB.pm,v 1.5 1996/01/18 17:47:24 olson Exp
#
# HDB.pm,v
# Revision 1.5  1996/01/18 17:47:24  olson
# added rcs headers
#
#

#
# Object: HDB
#
#
# Instance variables:
#
#	OBJ_config_file_entries
#
#	location
#
#		Directory on disk where databases are stored
#
#	hash_ref
#
#		Reference to tie'd hash.
#
#	hash
#
#		Hash which is tied to database.
#
#	next_new_host_index
#	next_search_host_index
#
# Methods
#
#	get_next_url	
#

#
# Keys
#
$K_index = 'index';
$K_host_for_index = 'host_for_index';
$K_next_url_index = 'next_url_index';

$max_queue = 500;
%host_in_use = ();
@walker_host = ();

@ISA = qw(Object);

sub boot
{
    Nexus::register_handler("add_url", "add_url",
			    method_call => 1, unpack_buffer => 1);
    Nexus::register_handler("get_url", "get_url",
			    method_call => 1, unpack_buffer => 1);
    Nexus::register_handler("synch", "synch",
                            method_call => 1, unpack_buffer => 1);
}

sub new
{
    my $self = {};
    bless $self;
    $$self{'OBJ_config_file_entries'} = {};

    $| = 1;

    $self->generate_accessors('location',
			      'hash_ref',
			      'hash',
			      'host_url_db_cache',
			      'walker_queue');

    $self->set_walker_queue([]);
    $self->set_host_url_db_cache({});

    return $self;
}

sub next_new_host_index
{
    my($self) = @_;
    my($idx);

    if (!defined($idx = $self->hash->{'next_new_host_index'}))
    {
	$idx = 0;
	$self->hash->{'next_new_host_index'} = $idx;
    }
    return $idx;
}

sub increment_new_host_index
{
    my($self) = @_;
    my($idx);

    if (!defined($idx = $self->hash->{'next_new_host_index'}))
    {
	$idx = 0;
    }

    $self->hash->{'next_new_host_index'} = $idx + 1;
    print "increment_new_host_index returns $idx\n" if $debug;
    return $idx;
}

sub increment_search_host_index
{
    my($self) = @_;
    my($idx, $nidx);

    if (!defined($idx = $self->hash->{'next_search_host_index'}))
    {
	print "incr: setting idx=0\n" if $debug;
	$idx = 0;
    }

    $nidx = $idx + 1;
    if ($nidx >= $self->next_new_host_index)
    {
	print "incr: wrapping nidx=$nidx to 0\n";
	$nidx = 0;
    }

    print "incr: setting next to $nidx returning $idx\n" if $debug;
    $self->hash->{'next_search_host_index'} = $nidx;

    return $idx;
}

sub add_new_host
{
    my($self, $host) = @_;
    my($idx);

    $idx = $self->increment_new_host_index();
    print "HDB: add_new_host host=$host idx=$idx\n" if $debug;
    $self->hash->{$host, $K_index} = $idx;
    $self->hash->{$K_host_for_index, $idx} = $host;
    $self->hash->{$host, $K_next_url_index} = 0;
    return $idx;
}

sub index_for_host
{
    my($self, $host) = @_;
    my($idx);

    if (!defined($idx = $self->hash->{$host, $K_index}))
    {
	$idx = $self->add_new_host($host);
    }
    return $idx;
}

sub host_for_index
{
    my($self, $idx) = @_;
    my($host);

    if (!defined($host = $self->hash->{$K_host_for_index, $idx}))
    {
	return undef;
    }
    return $host;
}

sub get_url
{
    my($self, $reply_gp, $reply_func, $id) = @_;
    my($url, $start_idx, $idx, $host, $db, $next_url);

    print "HDB $id get_url\n" if $debug_sched;

    if ($walker_host[$id] ne '')
    {
	print "HDB: freeing host $walker_host[$id] for walker $id\n" if $debug_sched;
	delete $host_in_use{$walker_host[$id]};
	undef $walker_host[$id];

    }
  
    $start_idx = $idx = $self->increment_search_host_index();
    $host = $self->host_for_index($start_idx);

    print "got idx=$idx\n" if $debug;
    while (1)
    {
	$host = $self->host_for_index($idx);
	if (!defined($host_in_use{$host}))
	{
	    $host_in_use{$host} = 1;
	    $next_url = $self->hash->{$host, $K_next_url_index};
  	    $db = $self->get_host_url_db($host);
	    print "get_url: host=$host next_url=$next_url db-len=", $db->length, "\n" if $debug;
            if ($next_url >= $db->length)
    	    {
		delete $host_in_use{$host};
	        $idx = $self->increment_search_host_index();
		print "No urls here, get next host\n" if $debug;
	        if ($idx == $start_idx)
	        {
	            push(@{$self->walker_queue}, [$reply_gp, $reply_func, $id]);
	            print "HDB: Walker $id placed on queue\n" if $debug_sched;
	            return undef;
	        }
	    }
	    else
	    {
	        $url = $db->FETCH($next_url);
	        $next_url++;
	        $self->hash->{$host, $K_next_url_index} = $next_url;
	        print "HDB: Got url $url for walker $id\n" if $debug_sched;
		#
		# Why would we do this?
		# undef($host_in_use{$host});
	        last;
	    }
        }
	else
	{
	    print "This host taken, get next one\n" if $debug;
	    $idx = $self->increment_search_host_index();
	    if ($idx == $start_idx)
	    {
		push(@{$self->walker_queue}, [$reply_gp, $reply_func, $id]);
	        print "HDB: Walker $id placed on queue\n" if $debug_sched;
		return undef;
	    }
	}
    }
#    print "HDB invoking <$reply_gp> <$reply_func> <$url>\n";
    print "HDB: scheduling $url for walker $id\n" if $debug_sched;
    $walker_host[$id] = $host;
    Nexus::rpc_async($reply_gp, $reply_func, $url);
}

#
# Start up the object.
#
sub start
{
    my($self) = @_;
    my($db_loc, $flags);
    my($idx);

    print "start self=$self loc=", $self->location, "\n" if $debug;
    if ($self->location eq '')
    {
	print "no location!\n" if $debug;
	return;
    }
    if (! -d $self->location)
    {
	mkdir($self->location, 0775);
    }

    $db_loc = $self->location . '/hash';

    $flags = 0;
    if (! -f $db_loc)
    {
	$flags |= O_CREAT;
    }
    $flags |= O_RDWR;

    $self->set_hash({});
    printf "loc=$db_loc flags=%o\n", $flags if $debug;
    $self->set_hash_ref(tie(%{$self->hash},
			    DB_File, $db_loc, $flags, 0664,
			    $DB_File::DB_HASH));
}

#
# Add the url to the url's host's list of urls.
#
sub add_url
{
    my($self, $url_string) = @_;
    my($db, $url);
    my($scheme,$host,$port,$path,$query,$frag);

    $url = new URI::URL $url_string;

    return unless $url;

    $scheme = $url->scheme;
    $host = $url->host; 
    $port = $url->port;
    $path = $url->path;
    $query = $url->query;
    $frag = $url->frag;

#    print "HDB: adding url $url_string for host $host\n";

    #
    # Force the host index to be initialized.
    #
    $self->index_for_host($host);

    $db = $self->get_host_url_db($host);

    $db->STORE($db->length, $url);
    print "HDB add_url $url\n" if $debug;

    $walker = pop(@{$self->walker_queue});
    if (defined($walker))
    {
        ($reply_gp, $reply_func, $id) = @$walker;
        print "HDB: dequeued walker $id\n" if $debug_sched;
        $self->get_url($reply_gp, $reply_func);
    }
}

#
# Retrieve the RECNO db ref for the given host.
#

sub get_host_url_db
{
    my($self, $host) = @_;
    my($db, $cache_entry, $cache);

    if (defined($cache_entry = $self->host_url_db_cache->{$host}))
    {
#	print "Here is the entire cache $cache_entry\n";
#	print "DB value -- Here is host $host and cache $$cache_entry[0]\n";
	return $$cache_entry[0];
    }

    $cache_entry = [undef, {}];

    if ($queue_val == $max_queue)
    {
	print "Filled queue\n" if $debug;
        $old_host = shift(@queue);
	$cache = $self->host_url_db_cache->{$old_host};
	untie(@$cache[1]);
        $self->host_url_db_cache->{$old_host} = undef;	
        $queue_val--;
    }

    ($db_loc, $flags) = $self->host_db_location($host);
    printf "Tying $host to $db_loc flags=%o\n", $flags if $debug;
#    print "DB location -- $db_loc\n"; 
    $db = tie(@{$cache_entry[1]}, DB_File,
			  $db_loc, $flags, 0664, $DB_File::DB_RECNO);

    $$cache_entry[0] = $db;
#    print "New DB -- Here is new host $host and new cache $$cache_entry[0]\n";
    $self->host_url_db_cache->{$host} = $cache_entry;
    push(@queue, $host);
    $queue_val++;
    $total++;
    print "Number in queue is $queue_val, total is $total\n" if $debug;
#    print "This is the entire cache $cache_entry\n";
    return $db;
}

sub host_db_location
{
    my($self, $host) = @_;
    my($db_loc);
    my(@h, $h);

    @h = reverse(split(/\./, $host));

    $db_loc = $self->location;

    for $h (@h)
    {
	$db_loc .= '/' . $h;
	if (! -d $db_loc)
	{
	    if (!mkdir($db_loc, 0775))
	    {
		warn "Cannot mkdir $db_loc: $!\n";
		return '/tmp';
	    }
	}
    }
    $db_loc .= '/url.recno';
    $flags = O_RDWR;
    $flags |= O_CREAT if ! -f $db_loc;
    return ($db_loc, $flags);
}

sub synch
{
    my($self) = @_;
    my($host, $cache_entry, $db);

    return unless ref($self->hash_ref());

    print "HDB Synch\n";

    $self->hash_ref()->sync;

    while (($host, $cache_entry) = each(%{$self->host_url_db_cache}))
    {
	if (ref($cache_entry))
	{
	    $db = $$cache_entry[0];
	    $db->sync();
	}
    }
}


sub quit
{
    my($self) = @_;
    my($host, $cache_entry, $db, $r);
    print "HDB quit\n";
    $self->synch();
    $r = untie %{$self->hash};
    print "HDB untie returns $r\n";
    while (($host, $cache_entry) = each(%{$self->host_url_db_cache}))
    {
	$db = $$cache_entry[1];
	print "Untie $db\n";
	untie %{$db};
    }
    Nexus::destroy_current_context();
}

sub DESTROY
{
    my($self) = @_;
    $self->quit();
}

1;
