#
#  Get a file using ftp.
#
#   ftpget host source dest
#      gets file, logging in as anonymous with user@host password (default)
#   ftpget user@host source dest
#      gets file as user "user" and prompts for password
#
#   if "dest" is a directory, filename is same as remote file name
#
#  Written by Gene Spafford  <spaf@cs.purdue.edu>
#   Last update, 17 May 1994  D. Sundstrom  <sunds@lobby.ti.com>
#
package ftpget;

require "ftplib2.pl";
require URI::URL;

#
# Retrieve Web page based on current url.
#
sub retrieve
{
    local($user, $pass, $url_string, *headers, *content, $timeout) = @_;
    my($scheme, $host, $port, $path, $query, $frag);

    $url = new URI::URL $url_string;

    $scheme = $url->scheme;
    $host = $url->host;
    $port = $url->port;
    $path = $url->path;
    $query = $url->query;
    $frag = $url->frag;

    $path = "/$path" unless $path =~ m:^/:;
    if (&ftp'open($host, $user, $pass))
    {
        if (&ftp'type($Ascii ? "a" : "i"))
	{
            &ftp'get($path, *content, $timeout) || &fail($url, "get");
        }
	else
	{
	    &fail($url, "type");
        }
    }
    else
    {
	&fail($url, "open");
    }
    &ftp'close;
}

#
# Failure occurred, print out cause.
#
sub fail
{
    my($url, $cause) = @_;
    my($save);

    $save = &ftp'error;
    &ftp'close;
    print "The url $url failed for $cause\n";
    return 0;
}

