package gopher_get; 

require "gchat.pl";
require URI::URL;

$gopher_port = 70;

#
# Split url in order to retrieve the Wab page connected to it.
#
sub retrieve
{
    local($url_string, *headers, *content, $timeout) = @_;
    my($scheme, $netloc, $port, $gtype, $selector, $search);

    $url = new URI::URL $url_string;

    $scheme = $url->scheme;
    $netloc = $url->netloc;
    $port = $url->port;
    if ($port == '')
    {
	$port = $gopher_port;
    }
    $path = $url->path;
    $path = "/$path" unless $path =~ m:^/:;
    $gtype = $1 if $path =~ s!^/(.)!!;
    $selector = $1 if $path =~ s!([^\?]*)!!;
    $search = $path;

#    if ($url =~ m#^\s*\w+://([\w-\.]+):?(\d*)/(\w?)([^ \t\?]*)\??(.*)$#)
#    {
#	$host = $1;
#	$host =~ tr/A-Z/a-z/;
#       $port = ($2 ne "" ? $2 : $gopher_port);
#	$gtype = ($3 ne "" ? $3 : 1);
#	$selector = $4;
#	$search = $5;
#    }
#    $gtype =~ s/%(\w\w)/sprintf("%c", hex($1))/ge; 
#    $selector =~ s/%(\w\w)/sprintf("%c", hex($1))/ge if ($selector);
#    $search =~ s/%(\w\w)/sprintf("%c", hex($1))/ge if ($search);

     &get($netloc, $port, $gtype, $selector, $search, *content, $timeout);
     if ($content ne "")
     {
	 return 2;
     }
}

#
# Connect to gopher site and retrieve Web page based on current url.
#
sub get
{
    local($host, $port, $gtype, $selector, $search, *content, $timeout) = @_;
    local($bintypes) = "59sgI";       # Binary gopher types
    local($goodtypes) = "01579sghI";  # types we can handle
    local($request, $handle, $done);

    $request = ($search ? "$selector\t$search\t\$" : $selector);
    if ($handle = &chat'open_port($host, $port))
    {
        if (&chat'print($handle, "$request\n"))
        { 
            *S = *chat'S;
        }
	else
	{
	    warn "chat'print($request): $!\n";
	    return(0);
        }
    }
    else
    {
	warn "chat'open($host,$port): $!\n";
	return(0);
    }

    if (index($goodtypes, $gtype) == -1)
    {
        warn "Can't retrieve gopher type $gtype\n";
	return(0);
    }

# If this is a binary document, retreive it using sysreads rather
# than <S>

    if (index($bintypes, $gtype) > -1)
    {
        $done = 0;
	$rmask = "";
	vec($rmask,fileno(S),1) = 1;
	do
	{
            ($nfound, $rmask) = select($rmask, undef, undef, $timeout); 
	    if ($nfound)
	    {
	        $nread = sysread(S, $thisbuf, 1024);
		if ($nread > 0)
		{
		    $content .= $thisbuf;
	            $content .= $thisbuf;
                }
		else
		{
	            $done++;
		}
	    }
	    else
	    {
		warn "Timeout\n"; $done++;
	    }
	} until $done;
    }

# This is an ASCII document, and we can get it line-by-line using <S>

    else
    {
        while (<S>)
	{
	    last if (/^\.\r\n$/);
	    chop;
	    chop;
	    $content .= "$_\n";
	}
    }
    &chat'close($handle);
    return;
}

